from restiro.helpers import to_snake_case, replace_non_alphabet
from .translation_mixin import TranslationMixin


class Document(TranslationMixin):
    __translation_keys__ = (
        'title',
        'content'
    )

    def __init__(self, title: str, content: str=None):
        """
        Additional documentation.
        The API definition can include a variety of documents that serve 
        as a user guides and reference documentation for the API. 
        Such documents can clarify how the API works or provide business context.
        Documentation-generators MUST include all the sections in an API definition's 
        documentation property in the documentation output, and they MUST preserve the 
        order in which the documentation is declared.

        :param title: The title of documentation.
        :param content: The content of documentation formatted in ``markdown``.
        """
        self.title, self.content = title, content

    @property
    def __filename__(self):
        return replace_non_alphabet(to_snake_case(self.title), '-')

    def to_dict(self):
        return {
            'title': self.title,
            'content': self.content
        }


class Documents(list):

    def append(self, document: Document):
        if not isinstance(document, Document):
            raise TypeError('item is not of type Document')
        super().append(document)

    def to_dict(self):
        return [document.to_dict() for document in self]

    def extract_translations(self):
        result = []
        for document in self:
            result.extend(document.extract_translations())
        return result

    def translate(self, translator):
        for document in self:
            document.translate(translator)




from os.path import join
from uuid import uuid4

from webtest import TestApp as WebtestApp

from restiro import (
    ResourceExample,
    ExampleRequest,
    ExampleResponse
)
from restiro.helpers import get_examples_dir


class TestApp(WebtestApp):

    def __init__(self, *args, examples_dir: str=None, **kwargs):
        self._examples_dir = examples_dir or get_examples_dir()
        self.doc = False
        self.force_doc = False
        self.requests_index = 0
        super().__init__(*args, **kwargs)

    def do_request(self, req, status=None, expect_errors=None):
        if not self.doc and not self.force_doc:
            return super().do_request(req=req, status=status,
                                      expect_errors=expect_errors)

        self.doc = False
        self.requests_index += 1

        # Fill example
        example_request = ExampleRequest(
            path=str(req.path),
            method=str(req.method).lower(),
            headers=dict(req.headers),
            body=str(req.as_text()),
            query_strings=dict(req.GET),
            form_params=dict(req.POST))

        response = super().do_request(
            req=req,
            status=status,
            expect_errors=expect_errors)

        example_response = ExampleResponse(
            status=response.status_int,
            body=str(response.text),
            headers=dict(response.headers),
            reason=response.status[3:].strip())

        example_filename = join(
            self._examples_dir,
            '%s-%s.pickle' % (self.requests_index, uuid4().hex))

        ResourceExample(
            request=example_request,
            response=example_response
        ).dump(
            example_filename
        )

        return response



import shutil

from os import makedirs
from os.path import exists, dirname, join

from restiro import DocumentationRoot, Resource, Document


# noinspection PyMethodMayBeStatic
class BaseGenerator:
    _files = []

    def __init__(self, docs_root: DocumentationRoot, destination_dir: str):
        self.destination_dir = destination_dir
        self.docs_root = docs_root

    def _ensure_file(self, filename: str):
        filename = join(self.destination_dir, filename)
        d = dirname(filename)
        if not exists(d):
            makedirs(d)

        if filename in self._files:
            return open(filename, 'a')
        else:
            f = open(filename, 'w')
            self._files.append(filename)
            return f

    def generate_resources(self):
        resources_tree = self.docs_root.resources.__tree__.items()
        for resource_path, resource_methods in resources_tree:
            for resource_method, resources in resource_methods.items():
                for resource in resources:
                    f = self._ensure_file(self.get_resource_filename(resource))
                    self.write_resource(f, resource)
                    f.close()

    def generate_documents(self):
        for document in self.docs_root.documents:
            f = self._ensure_file(self.get_document_filename(document))
            self.write_document(f, document)
            f.close()

    def generate_index(self):
        f = self._ensure_file(self.get_index_filename())
        self.write_index(f)
        f.close()

    def clean_destination(self):
        shutil.rmtree(self.destination_dir)

    def get_resource_filename(self, resource: Resource):
        return "%s" % resource.__filename__

    def get_document_filename(self, document: Document):
        return "-%s" % document.__filename__

    def get_index_filename(self):
        return 'index'

    def write_document(self, file_stream,
                       document: Document):  # pragma: nocover
        raise NotImplementedError

    def write_resource(self, file_stream,
                       resource: Resource):  # pragma: nocover
        raise NotImplementedError

    def write_index(self, file_stream):  # pragma: nocover
        raise NotImplementedError

    def generate(self):
        self.clean_destination()
        self.generate_documents()
        self.generate_resources()
        self.generate_index()




class DocstringApiDefinition:
    name = None
    title = None
    description = ''
    content = ''

    def __init__(self, docstring):
        for line in docstring.split('\n'):
            exploded_line = line.split(' ')
            if line.startswith('@apiDefine '):
                self.name = exploded_line[1]
                self.title = exploded_line[2] \
                    if len(exploded_line) > 2 else None
                self.description = ' '.join(exploded_line[3:]) \
                    if len(exploded_line) > 3 else ''

            elif line.startswith(' '):
                self.description += ' %s' % line.lstrip()

            else:
                self.content += line + '\n'



"""
    API Documentation models, mostly based on RAML 0.8
"""
from .parameters import URLParam, QueryParam, HeaderParam, FormParam, Param
from .example import (
    ResourceExample,
    ExampleResponse,
    ExampleRequest,
    BodyFormat,
    BodyFormatJson,
    BodyFormatYaml,
    BodyFormatXml
)
from .resource import Resource, Resources
from .document import Document
from .root import DocumentationRoot



from types import GeneratorType
from typing import List, Union, Generator

from .parameters import URLParam, FormParam, HeaderParam, QueryParam, Param
from .example import ResourceExample
from .translation_mixin import TranslationMixin


class Resource(TranslationMixin):
    __translation_keys__ = (
        'description',
        'display_name'
    )

    def __init__(self, path: str, method: str, display_name: str = None,
                 description: str = None, tags: List[str] = None,
                 params: Union[Param, List[Param], Generator] = None,
                 security: dict = None,
                 examples: Union[List[ResourceExample], Generator] = None):
        """
        Resource
        :param path: The URI relative to the `DocumentationRoot.base_uri` and 
                     MUST begin with a slash (/).
                     .. seealso:: :class:`.DocumentationRoot`
        :param method: The method of resource. In a RESTful API, methods are
                       operations that are performed on a resource.
                       A method MUST be one of the HTTP methods defined in
                       the HTTP version 1.1 specification [RFC2616] and its
                       extension, RFC5789 [RFC5789].
        :param display_name: The displayName attribute provides a friendly name
                             to the resource and can be used by documentation
                             generation tools.
        :param description: The description property that briefly describes
                            the resource. It is RECOMMENDED that all the API
                            definition's resources includes the description
                            property.
        :param params: Collection of parameters (Include any inheritance
                       of `Param`)
        :param security: Collection of permissions
        :param examples: Collection of examples
        """
        self.path = path
        self.method = method
        self.display_name = display_name
        self.description = description
        self.tags = tags
        self.uri_params = []
        self.query_params = []
        self.form_params = []
        self.security = security
        self.header_params = []
        self.examples = examples if examples else []

        if params:
            if isinstance(params, list):
                self.set_params(*params)
            elif isinstance(params, GeneratorType):
                self.set_params(*list(params))
            else:
                self.set_params(params)

    def set_params(self, *args):
        for item in args:
            if isinstance(item, URLParam):
                self.uri_params.append(item)

            elif isinstance(item, QueryParam):
                self.query_params.append(item)

            elif isinstance(item, FormParam):
                self.form_params.append(item)

            elif isinstance(item, HeaderParam):
                self.header_params.append(item)
        return self

    @property
    def params(self):
        """ List of all parameters """
        return (
            self.uri_params +
            self.query_params +
            self.header_params +
            self.form_params
        )

    @property
    def __key__(self):
        return '%s-%s' % (self.path, self.method)

    @property
    def __filename__(self):
        return str(self.__key__).lstrip('/').replace('/', '-')

    def to_dict(self):
        return {
            'path': self.path,
            'method': self.method,
            'tags': self.tags,
            'display_name': self.display_name,
            'description': self.description,
            'header_params': [param.to_dict() for param in self.header_params],
            'uri_params': [param.to_dict() for param in self.uri_params],
            'query_params': [param.to_dict() for param in self.query_params],
            'form_params': [param.to_dict() for param in self.form_params],
            'examples': [example.to_dict() for example in self.examples]
        }

    def __repr__(self):
        return '%s %s' % (self.method.upper(), self.path)

    def extract_translations(self):
        result = super().extract_translations()
        for param in self.params:
            result.extend(param.extract_translations())
        return result


class Resources(dict):

    def append(self, resource: Resource):
        if not isinstance(resource, Resource):
            raise TypeError('item is not of type Resource')
        self[resource.__key__] = resource

    def find(self, path, method) -> Resource:
        input_path_parts = path[1:].split('/')
        filtered_resources = [
            x for x in self.values()
            if (
                x.method == method and
                len(x.path[1:].split('/')) == len(input_path_parts)
            )
        ]

        def _route(resources, part_index=0):
            matched_resources = []
            for resource in resources:
                resource_path_parts = resource.path[1:].split('/')
                if (
                    input_path_parts[part_index] ==
                        resource_path_parts[part_index] or
                        resource_path_parts[part_index][:1] == ':'
                ):
                    matched_resources.append(resource)

            if part_index == len(input_path_parts) - 1:
                return matched_resources[0] if matched_resources else None

            return _route(matched_resources, part_index + 1)

        return _route(resources=filtered_resources)

    @property
    def __tree__(self):
        """
        Get tree of resources
        First level: resource
        Second level: method
        :return: 
        """
        result = {}
        for resource_key, resource in self.items():
            result.setdefault(resource.path, {})
            result[resource.path].setdefault(resource.method, [])
            result[resource.path][resource.method].append(resource)
        return result

    def to_dict(self):
        return [resource.to_dict() for resource_key, resource in self.items()]

    def extract_translations(self):
        result = []
        for resource in self.values():
            result.extend(resource.extract_translations())
        return result

    def translate(self, translator):
        for resource in self.values():
            resource.translate(translator)





class TranslationMixin:
    __translation_keys__: tuple = ()

    def extract_translations(self):
        return [
            getattr(self, key)
            for key in self.__translation_keys__
            if getattr(self, key) is not None
        ]

    def translate(self, translator):
        for key in self.__translation_keys__:
            value = getattr(self, key)
            if value is None:
                continue
            setattr(self, key, translator(value))



from typing import List

from restiro import (
    Resource,
    Document,
    ExampleResponse,
    ExampleRequest,
    ResourceExample)
from restiro.models.parameters import Param
from restiro.generators import BaseGenerator


class MarkdownRepresentations:

    @staticmethod
    def _repr_response_body(response: ExampleResponse):
        syntax_language = response.body_format
        return '```%s\n%s\n```' % (
            syntax_language.name if syntax_language else '',
            response.formatted_body
        )

    @staticmethod
    def _repr_request_headers(request: ExampleRequest):
        return '```\n%s\n```' % (
            request.text.split('\r\n\r\n')[0]
        )

    @staticmethod
    def _repr_response_headers(response: ExampleResponse):
        return '```\n%s\n```' % (
            response.repr_headers()
        )

    @staticmethod
    def _repr_request_body(request: ExampleRequest):
        syntax_language = request.body_format
        return '```%s\n%s\n```' % (
            syntax_language.name if syntax_language else '',
            request.body_text
        )

    @classmethod
    def _repr_example(cls, example: ResourceExample):
        """ Beautify and return example """
        return (
           '#### Request: \n\n'
           '%s\n'  # Headers
           '%s\n'  # Body
           '\n'
           '#### Response: \n\n'
           '%s\n'  # Headers
           '%s\n'  # Body
        ) % (
           cls._repr_request_headers(example.request),
           cls._repr_request_body(example.request),
           cls._repr_response_headers(example.response),
           cls._repr_response_body(example.response),
        )


class MarkdownGenerator(BaseGenerator, MarkdownRepresentations):

    def get_document_filename(self, document: Document):
        return '%s.md' % super().get_document_filename(document)

    def get_resource_filename(self, resource: Resource):
        return '%s.md' % super().get_resource_filename(resource)

    def get_index_filename(self):
        return '%s.md' % super().get_index_filename()

    def write_document(self, f, document: Document):
        f.write('# %s\n\n' % document.title)
        f.write('%s' % document.content)

    # noinspection PyMethodMayBeStatic
    def write_params(self, f, params: List[Param]):
        params_heads = (
            'Name', 'Type', 'Required', 'Default', 'Example', 'enum',
            'Pattern', 'MinLength', 'MaxLength', 'Minimum', 'Maximum',
            'Repeat', 'Description'
        )
        f.write('%s\n' % ' | '.join(params_heads))
        f.write('%s\n' % ' | '.join(list(map(lambda x: '---', params_heads))))
        value_placeholder = ' | '.join(list(map(lambda x: '%s', params_heads)))
        value_placeholder = '%s\n' % value_placeholder
        params.sort(key=lambda x: x.name)
        for param in params:
            f.write(
                value_placeholder % (
                    param.name,
                    param.type_,
                    '`True`' if param.required else '`False`',
                    param.default or '',
                    param.example or '',
                    param.enum or '',
                    param.pattern or '',
                    param.min_length or '',
                    param.max_length or '',
                    param.minimum or '',
                    param.maximum or '',
                    '`True`' if param.repeat else '`False`',
                    (param.description or '').replace('\n', ' ')
                )
            )

    # noinspection PyMethodMayBeStatic
    def write_resource_parameters(self, f, resource: Resource):

        if len(resource.uri_params) > 0:
            f.write('### URI parameters\n\n')
            self.write_params(f, resource.uri_params)
            f.write('\n')

        if len(resource.query_params) > 0:
            f.write('### Query-string parameters\n\n')
            self.write_params(f, resource.query_params)
            f.write('\n')

        if len(resource.header_params) > 0:
            f.write('### Header parameters\n\n')
            self.write_params(f, resource.header_params)
            f.write('\n')

        if len(resource.form_params) > 0:
            f.write('### Form parameters\n\n')
            self.write_params(f, resource.form_params)
            f.write('\n')

    def write_resource(self, f, resource: Resource):
        f.write('# %s\n\n' % (resource.display_name or 'untitled'))
        f.write('## `%s` `%s`\n\n' % (str(resource.method).upper(),
                                      resource.path))

        if resource.description:
            f.write('%s\n\n' % resource.description)

        if len(resource.params) > 0:
            f.write('\n## Parameters\n\n')
            self.write_resource_parameters(f, resource)

        if len(resource.examples) > 0:
            f.write('## Examples\n\n')
            for index, example in enumerate(resource.examples):
                f.write('### #%s\n\n' % index)
                f.write(self._repr_example(example))
                f.write('\n---\n')

    def write_index(self, file_stream):
        file_stream.write(
            '# %s\n\n' % self.docs_root.title or 'untitled'
        )

        if len(self.docs_root.documents) > 0:
            file_stream.write('\n## Documents\n\n')
            for document in self.docs_root.documents:
                file_stream.write('- %s\n' % document.title)

        if len(self.docs_root.resources) > 0:
            file_stream.write('\n## Resources\n\n')
            for resource_key, resource in self.docs_root.resources.items():
                file_stream.write(
                    '- `%s` `%s` \n' % (resource.path, resource.method)
                )



from typing import List, Union
from urllib.parse import urlparse
from os import scandir

from restiro.helpers import get_examples_dir
from .resource import Resource, Resources
from .document import Document, Documents
from .translation_mixin import TranslationMixin


class DocumentationRoot(TranslationMixin):
    __translation_keys__ = (
        'title',
    )

    def __init__(self, title: str, base_uri: str = None, locale: str=None,
                 version: str=None,
                 resources: Union[Resource, List[Resource]] = None,
                 documents: Union[Document, List[Document]] = None):
        """
        Documentation root class 
        :param title: The ``title`` property is a short plain text description of the RESTful API.
                      The ``title`` property's value SHOULD be suitable for use as a title for
                      the contained user documentation.

        :param base_uri: A RESTful API's resources are defined relative to the API's base URI.
                         The use of the ``base_uri`` field is OPTIONAL to allow describing APIs that
                         have not yet been implemented.
                         After the API is implemented (even a mock implementation) and can be accessed
                         at a service endpoint, the API definition MUST contain a ``base_uri`` property. 
                         The ``base_uri`` property's value MUST conform to the URI specification [RFC2396] 
                         or a Level 1 Template URI as defined in RFC 6570 [RFC6570].
                         The ``base_uri`` property SHOULD only be used as a reference value. 

        :param resources: List of the resources.
        :param documents: List of additional documentations.
        """
        self.title = title
        self.locale = locale
        self.version = version
        self.base_uri = base_uri
        self.documents = Documents()
        self.resources = Resources()

        if documents:
            if isinstance(documents, list):
                self.set_documents(*documents)
            else:
                self.set_documents(documents)

        if resources:
            if isinstance(resources, list):
                self.set_resources(*resources)
            else:
                self.set_resources(resources)

    def set_resources(self, *args):
        for item in args:
            self.resources.append(item)
        return self

    def set_documents(self, *args):
        for item in args:
            self.documents.append(item)
        return self

    @property
    def base_uri_path(self):
        return urlparse(self.base_uri).path if self.base_uri else ''

    def to_dict(self):
        return {
            'title': self.title,
            'locale': self.locale,
            'version': self.version,
            'base_uri': self.base_uri,
            'documents': self.documents.to_dict(),
            'resources': self.resources.to_dict(),
        }

    def extract_translations(self):
        result = super().extract_translations()
        result.extend(self.documents.extract_translations())
        result.extend(self.resources.extract_translations())
        return sorted(result)

    def translate(self, translator):
        super().translate(translator)
        self.documents.translate(translator)
        self.resources.translate(translator)

    def translate_all(self, locales_dir, locale, domain: str = 'restiro'):
        import gettext
        import locale as lib_locale

        translation = gettext.translation(
            domain=domain,
            localedir=locales_dir,
            languages=[locale]
        )
        try:
            lib_locale.setlocale(lib_locale.LC_ALL, locale)
        except lib_locale.Error:
            print('Invalid locale %s' % locale)
            raise

        self.translate(translation.gettext)

    def load_resource_examples(self, examples_dir: str=None):
        """
        Load example objects into resources

        :param examples_dir:
        :return:
        """
        from . import ResourceExample
        if not examples_dir:
            examples_dir = get_examples_dir()

        # Sort examples by index number from filename (generated by middleware)
        # Examples filename format: {index}-{uuid}.pickle
        directories = sorted(scandir(examples_dir),
                             key=lambda k: int(k.name.split('-')[0]))
        for dir_entry in directories:
            pickle_filename = dir_entry.path
            resource_example = ResourceExample.load(pickle_filename)

            resource_path = resource_example.request.path
            resource_method = resource_example.request.method

            resource = self.resources.find(
                path=resource_path,
                method=resource_method
            )

            if not resource:
                resource_path = resource_path[len(self.base_uri_path):]

                resource = self.resources.find(
                    path=resource_path,
                    method=resource_method
                )

            if not resource:
                continue

            resource.examples.append(resource_example)



from .models import (
    Resources, Resource, ResourceExample, ExampleRequest, ExampleResponse,
    DocumentationRoot, Document,
    Param, URLParam, QueryParam, HeaderParam, FormParam,
    BodyFormat, BodyFormatJson, BodyFormatYaml, BodyFormatXml
)
from .parser import Parser
from .documentor import Documentor
from .helpers import clean_examples_dir

__version__ = '0.9.0'



import re

docstring_block_regex = re.compile(r'\"\"\"([\s\S]*?)\"\"\"')
within_parentheses_regex = re.compile(r'\(([\s\S]*?)\)')
within_brackets_regex = re.compile(r'{([\s\S]*?)}')
single_word_regex = re.compile(r'\s(\[?\w+\]?)(?=\s?)')
path_regex = re.compile(r'\s/([/\w:]+)(?=\s?)')



import json
import pickle

from typing import Union

from restiro.helpers import CaseInsensitiveDict


class BodyFormat:
    name = None
    header_mime = None


class BodyFormatJson(BodyFormat):
    name = 'json'
    header_mime = 'application/json'


class BodyFormatText(BodyFormat):
    name = 'text'
    header_mime = 'text/plain'


class BodyFormatXml(BodyFormat):
    name = 'xml'
    header_mime = 'application/xml'


class BodyFormatYaml(BodyFormat):
    name = 'yaml'
    header_mime = 'application/x-yaml'


class ExampleRequest:
    def __init__(self, path: str, method: str, headers: dict = None,
                 query_strings: dict = None, form_params: dict = None,
                 body: str = None):
        self.path = path
        self.method = method
        self.headers = dict(map(
            lambda x: (x[0].lower(), x[1]),
            headers.items()
        )) if headers else None
        self.query_strings = query_strings
        self.form_params = form_params
        self.body = body

    @property
    def body_format(self) -> Union[BodyFormat, None]:
        content_type_raw = CaseInsensitiveDict(self.headers).\
            get('Content-Type', None)
        return {
            BodyFormatJson.header_mime: BodyFormatJson,
            BodyFormatYaml.header_mime: BodyFormatYaml,
            BodyFormatXml.header_mime: BodyFormatXml
        }.get(content_type_raw.split(';', 1)[0], None) \
            if content_type_raw else None

    @property
    def body_text(self):
        if self.body is None:
            return ''

        parsed_text = self.body.split('\r\n\r\n')
        return '\r\n\r\n'.join(parsed_text[1:]) \
            if len(parsed_text) > 1 else ''

    def __repr__(self):
        return self.body_text

    def to_dict(self):
        return {
            'path': self.path,
            'method': self.method,
            'headers': self.headers if self.headers else None,
            'query_strings': self.query_strings if self.query_strings else None,
            'form_params': self.form_params if self.form_params else None,
            'body_format': self.body_format.name if self.body_format else None,
            'body_text': self.body_text
        }


class ExampleResponse:

    def __init__(self, status: int, headers: dict, body: str, reason: str=None):
        self.status = status
        self.headers = dict(map(
            lambda x: (x[0].lower(), x[1]),
            headers.items()
        ))
        self.body = body
        self.reason = reason

    @property
    def body_format(self) -> Union[BodyFormat, None]:
        content_type_raw = CaseInsensitiveDict(self.headers).\
            get('Content-Type', None)
        return {
            BodyFormatJson.header_mime: BodyFormatJson,
            BodyFormatYaml.header_mime: BodyFormatYaml,
            BodyFormatXml.header_mime: BodyFormatXml
        }.get(content_type_raw.split(';', 1)[0], None) \
            if content_type_raw else None

    @property
    def body_json(self):
        return json.dumps(self.body, indent=4)

    def to_dict(self):
        return {
            'status': self.status,
            'reason': self.reason,
            'headers': self.headers,
            'body_format': self.body_format.name if self.body_format else None,
            'body': self.body
        }

    def repr_headers(self):
        return '\n'.join('%s: %s' % header for header in self.headers.items())

    def __repr__(self):
        return '%s%s' % (
            self.repr_headers(),
            '\r\n\r\n%s' % self.body
        )


class ResourceExample:
    def __init__(self, request: ExampleRequest, response: ExampleResponse):
        self.request = request
        self.response = response

    def to_dict(self):
        return {
            'request': self.request.to_dict(),
            'response': self.response.to_dict()
        }

    def dump(self, pickle_filename):
        with open(pickle_filename, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, pickle_filename) -> 'ResourceExample':
        with open(pickle_filename, 'rb') as f:
            return pickle.load(f)



import argparse
from importlib.util import find_spec

from os import scandir, makedirs
from os.path import dirname, isdir, basename, join

from restiro import Documentor
from restiro.helpers import validate_locale_name

parser = argparse.ArgumentParser(description='Restiro Builder')
parser.add_argument(
    'src', help='Project module name')
parser.add_argument(
    '-t', '--title', help='Project title')
parser.add_argument(
    '-o', '--output', default='./docs', help='Output directory')
parser.add_argument(
    '-b', '--base-uri', help='Base URI')
parser.add_argument(
    '-g', '--generator', choices=('markdown', 'json', 'spa_material'),
    default='markdown', help='Generator, default: markdown')
parser.add_argument(
    '-l', '--locales', default='./locales', help='Locales directory')
parser.add_argument(
    '--build-gettext', default=False, const=True, nargs='?',
    help='Build .POT templates')


def main():
    args = parser.parse_args()
    title = args.title or args.src
    source_dir = dirname(find_spec(args.src).origin)
    locales_dir = args.locales
    output_base_dir = args.output

    def get_documentor():
        return Documentor(
            title=title,
            base_uri=args.base_uri,
            source_dir=source_dir,
            generator_type=args.generator
        )

    if args.build_gettext:
        gettext_dir = join(locales_dir, 'gettext')
        makedirs(gettext_dir, exist_ok=True)

        get_documentor().generate_gettext(gettext_dir)
        print('gettext .POT templates generated in: %s' % gettext_dir)
        return

    if isdir(locales_dir):
        for entry in scandir(locales_dir):
            if not entry.is_dir():
                continue

            locale = basename(entry.path)

            if not validate_locale_name(locale):
                continue

            output_dir = join(output_base_dir, locale)
            makedirs(output_dir, exist_ok=True)
            get_documentor().generate(output_dir, locales_dir, locale)

            print('Documentation build success. (%s)' % output_dir)

    else:
        get_documentor().generate(output_base_dir)
        print('Documentation build success. (%s)' % output_base_dir)




import re
import glob
import textwrap

from restiro.models import Resources
from restiro.constants import docstring_block_regex
from restiro.parser.resource import DocstringApiResource
from restiro.parser.definition import DocstringApiDefinition


class DocstringParser:

    def parse_docstring(self, docstring, filename, start_line):
        raise NotImplementedError

    def load_from_path(self, base_path: str = '.'):
        """ Load python files  """
        for filename in glob.iglob('%s/**/*.py' % base_path, recursive=True):
            self.load_file(filename)

    def load_file(self, filename: str):
        """ Open python file and parse docstrings """
        with open(filename, 'r') as f:
            source = f.read()
            docstring_blocks = self.find_docstring_blocks(source)
            for docstring_block in docstring_blocks:
                self.parse_docstring(docstring_block[0], filename,
                                     docstring_block[1])

    @staticmethod
    def find_docstring_blocks(source):
        """ Find docstring blocks from python source
            and return them with line number"""

        re_newline = re.compile(r'\n')
        all_doc_blocks = []
        for line in re.finditer(docstring_block_regex, source):
            start_line = len(re_newline.findall(source, 0, line.start())) + 1
            all_doc_blocks.append((
                textwrap.dedent(line.group()[3:-3]).lstrip(),
                start_line))

        return all_doc_blocks


class DocstringResourceParser(DocstringParser):

    def __init__(self, definitions: dict):
        self.resources = []
        self.definitions = definitions

    def parse_docstring(self, docstring, filename, start_line):
        if docstring.startswith('@api '):
            self.resources.append(
                DocstringApiResource(docstring,
                                     filename=filename,
                                     start_line=start_line,
                                     definitions=self.definitions)
            )

    def export_to_model(self) -> Resources:
        result = Resources()
        for resource in self.resources:
            result.append(resource.to_model())
        return result


class DocstringDefinitionParser(DocstringParser):

    def __init__(self):
        self.definitions = {}

    def parse_docstring(self, docstring, filename, start_line):
        if docstring.startswith('@apiDefine '):
            definition = DocstringApiDefinition(docstring)
            self.definitions[definition.name] = definition




import re
import collections
import tempfile

from os import makedirs
from os.path import join
from shutil import rmtree

_under_scorer1 = re.compile(r'(.)([A-Z][a-z]+)')
_under_scorer2 = re.compile('([a-z0-9])([A-Z])')
_non_alphabet = re.compile('[\x00-\x2F\x3A-\x40\x5B-\x60\x7B-\x7F]+')


def to_snake_case(s):
    """
    Convert camelCase to snake_case
    """
    subbed = _under_scorer1.sub(r'\1_\2', s)
    return _under_scorer2.sub(r'\1_\2', subbed).lower()


def replace_non_alphabet(string, replace):
    """
    Replace non-alphabetic characters
    """
    return _non_alphabet.sub(replace, string)


def get_examples_dir():
    temp_dir = join(tempfile.gettempdir(), 'restiro_examples')
    makedirs(temp_dir, exist_ok=True)
    return temp_dir


def clean_examples_dir():
    rmtree(get_examples_dir())


def generate_pot(translations):
    return ''.join(['msgid "%s"\nmsgstr ""\n\n' % t for t in translations])


def validate_locale_name(locale):
    return bool(re.fullmatch(r'^[a-z]{2,3}_[A-Z]{2}$', locale))


class CaseInsensitiveDict(collections.MutableMapping):
    """A case-insensitive ``dict``-like object.

    Implements all methods and operations of
    ``collections.MutableMapping`` as well as dict's ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive::

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.
    """

    def __init__(self, data=None, **kwargs):
        self._store = collections.OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        # Use the lower-cased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (cased_key for cased_key, mapped_value in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lower_key, key_val[1])
            for (lower_key, key_val)
            in self._store.items()
        )

    def __eq__(self, other):
        if isinstance(other, collections.Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self):
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))



from typing import Type

from os.path import join
from importlib import import_module

from restiro import Parser, DocumentationRoot
from restiro.helpers import generate_pot
from restiro.generators import BaseGenerator


class Documentor:

    def __init__(self, title: str, source_dir: str, base_uri: str=None,
                 generator_type: str='markdown'):
        self.title = title
        self.source_dir = source_dir
        self.base_uri = base_uri
        self.generator_type = generator_type

    def initiate_docs_root(self, locale=None):
        parsed_resources = Parser.load_from_path(self.source_dir)
        docs_root = DocumentationRoot(
            title=self.title,
            base_uri=self.base_uri,
            locale=locale
        )
        docs_root.resources.update(parsed_resources)
        docs_root.load_resource_examples()
        return docs_root

    @property
    def generator(self) -> Type[BaseGenerator]:
        try:
            module_name = (
                'restiro.generators.%s' % self.generator_type
                if self.generator_type in ('json', 'markdown') else
                'restiro_%s' % self.generator_type
            )
            mod = import_module(module_name)
            for cname in dir(mod):
                if not cname.endswith('Generator') or cname == 'BaseGenerator':
                    continue

                return getattr(mod, cname)

        except ImportError:
            raise ValueError('Generator not detected %s' % self.generator_type)

    def generate(self, output_dir: str, locales_dir=None, locale=None):
        docs_root = self.initiate_docs_root(locale)
        if locale:
            docs_root.translate_all(locales_dir, locale)

        self.generator(
            docs_root=docs_root,
            destination_dir=output_dir
        ).generate()

    def generate_gettext(self, gettext_dir):
        docs_root = self.initiate_docs_root()
        pot_file = join(gettext_dir, 'restiro.pot')
        with open(pot_file, 'w+') as f:
            f.write(generate_pot(docs_root.extract_translations()))




from restiro import Resources
from .docstring import DocstringDefinitionParser, DocstringResourceParser


class Parser:

    @staticmethod
    def load_from_path(base_path: str = '.') -> Resources:
        """
        Load and parse files 
        
        :param base_path: 
        :return: List of resources that grouped by version
        """
        definition_parser = DocstringDefinitionParser()
        definition_parser.load_from_path(base_path)
        resource_parser = DocstringResourceParser(definition_parser.definitions)
        resource_parser.load_from_path(base_path)
        return resource_parser.export_to_model()



from .translation_mixin import TranslationMixin

python_type_alias = {
    'int': 'integer',
    'str': 'string',
    'bool': 'boolean',
    'Decimal': 'number',
    'datetime': 'date',
    'date': 'date',
    'float': 'number',
    'TextIOWrapper': 'file',
    'TextIOBase': 'file',
    'BufferedIOBase': 'file',
    'RawIOBase': 'file',
    'IOBase': 'file',
    'IO': 'file',
    'FileIO': 'file',
    'BytesIO': 'file',
    'bytes': 'file',
}


class Param(TranslationMixin):
    __translation_keys__ = (
        'description',
        'display_name'
    )

    def __init__(self, name: str = None, display_name: str = None,
                 description: str = None, type_=None, enum: list = None,
                 pattern: str = None, min_length: int = None,
                 max_length: int = None, minimum: int = None,
                 maximum: int = None, example: str = None, repeat: bool = False,
                 required: bool = False, default: str = None):
        """
        Parameters base class

        :param name: Parameter name
        :param display_name: The ``display_name`` attribute specifies the
                             parameter's display name. It is a friendly name
                             used only for display or documentation purposes.
                             If ``display_name`` is not specified, it defaults
                             to the property's key (the name of
                             the property itself).
        :param description: The description attribute describes the intended
                            use or meaning of the parameter. This value MAY be
                            formatted using Markdown [MARKDOWN].
        :param type_: The ``type_`` attribute specifies the primitive 
                      type of the parameter's resolved value. API clients MUST
                      return/throw an error
                      if the parameter's resolved value does not match the
                      specified type.
                      If type is not specified, it defaults to string.
                      Valid types are:
                      string    Value MUST be a string.
                      number    Value MUST be a number.
                      integer   Value MUST be an integer. 
                                Floating point numbers are not allowed. 
                                The integer type is a subset of the number type.
                      date      Value MUST be a string representation of a date
                                as defined in RFC2616 Section 3.3 [RFC2616].
                                As defined in [RFC2616], all date/time stamps
                                are represented in Greenwich Mean Time (GMT),
                                which for the purposes of HTTP is equal to UTC
                                (Coordinated Universal Time).
                                This is indicated by including "GMT" as the
                                three-letter abbreviation for the timezone.
                                Example: Sun, 06 Nov 1994 08:49:37 GMT.
                      boolean   Value MUST be either the string "true" or
                                "false" (without the quotes).
                      file      (Applicable only to Form properties)
                                Value is a file. Client generators SHOULD use 
                                this type to handle file uploads correctly.
        :param enum: The ``enum`` attribute provides an enumeration of the
                     parameter's valid values.This MUST be an array.
                     If the ``enum`` attribute is defined, API clients and
                     servers MUST verify that a parameter's value matches a
                     value in the enum array.
                     If there is no matching value, the clients and servers
                     MUST treat this as an error.
                     (Applicable only for parameters of type string)
        :param pattern: The ``pattern`` attribute is a regular expression 
                        that a parameter of type string MUST match. Regular
                        expressions MUST follow the regular expression
                        specification from ECMA 262/Perl 5.
                        The pattern MAY be enclosed in double quotes for
                        readability and clarity.
                        (Applicable only for parameters of type string) 
        :param min_length: The ``min_length`` attribute specifies the parameter
                           value's minimum number of characters.
                           (Applicable only for parameters of type string)
        :param max_length: The ``max_length`` attribute specifies the parameter
                           value's maximum number of characters.
                           (Applicable only for parameters of type string)
        :param minimum: The ``minimum`` attribute specifies the parameter's
                        minimum value. (Applicable only for parameters of type
                        number or integer)
        :param maximum: The ``maximum`` attribute specifies the parameter's
                        maximum value. (Applicable only for parameters
                        of type number or integer)
        :param example: The ``example`` attribute shows an example value for
                        the property. This can be used, e.g., by documentation
                        generators to generate sample values for the property.
        :param repeat: The ``repeat`` attribute specifies that the parameter
                       can be repeated. If the parameter can be used multiple
                       times, the repeat parameter value MUST be set to 'true'.
                       Otherwise, the default value is 'false' and the
                       parameter may not be repeated.
        :param required: The ``required`` attribute specifies whether the
                         parameter and its value MUST be present in the API
                         definition. It must be either 'true' if the value MUST
                         be present or 'false' otherwise.
                         In general, parameters are optional unless the
                         required attribute is included and its value set to
                         'true'.
                         For a URI parameter, the required attribute MAY be
                         omitted, but its default value is 'true'.
        :param default: The ``default`` attribute specifies the default value
                        to use for the property if the property is omitted or
                        its value is not specified.
                        This SHOULD NOT be interpreted as a requirement for the
                        client to send the default attribute's value if there
                        is no other value to send. Instead, the default
                        attribute's value is the value the server uses if the
                        client does not send a value.
        """
        self.name = name
        self.display_name = display_name
        self.description = description
        self.type_ = type_ if isinstance(type_, str) else \
            self.get_type_string(type_)
        self.enum = enum
        self.pattern = pattern
        self.min_length = min_length
        self.max_length = max_length
        self.minimum = minimum
        self.maximum = maximum
        self.example = example
        self.repeat = repeat
        self.required = required
        self.default = default

    @staticmethod
    def get_type_string(type_):
        if hasattr(type_, '__name__'):
            return python_type_alias.get(type_.__name__, None)

    def to_dict(self):
        return {
            'name': self.name,
            'display_name': self.display_name,
            'description': self.description,
            'type': self.type_,
            'enum': self.enum,
            'pattern': self.pattern,
            'min_length': self.min_length,
            'max_length': self.max_length,
            'minimum': self.minimum,
            'maximum': self.maximum,
            'example': self.example,
            'repeat': self.repeat,
            'required': self.required,
            'default': self.default
        }


class URLParam(Param):
    def __init__(self, *args, required: bool = True, **kwargs):
        super().__init__(*args, required=required, **kwargs)


class QueryParam(Param):
    pass


class FormParam(Param):
    pass


class HeaderParam(Param):
    pass



import re

from os.path import dirname, join
from setuptools import setup, find_packages


# Read package version
with open(join(dirname(__file__), 'restiro', '__init__.py')) as f:
    package_version = re.compile(r".*__version__ = '(.*?)'",
                                 re.S).match(f.read()).group(1)

setup(
    name='restiro',
    version=package_version,
    description='RESTful API documentation generator',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='http://github.com/meyt/restiro',
    author='Mahdi Ghane.g',
    license='GPLv3',
    keywords='api_documentation api document_generator',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Documentation',
        'Topic :: Documentation',
        'Topic :: Software Development :: Build Tools'
    ],
    packages=find_packages(),
    install_requires=[],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'restiro = restiro.cli:main'
        ]
    }
)



import json

from restiro.generators import BaseGenerator


class JSONGenerator(BaseGenerator):

    def get_index_filename(self):
        return '%s.json' % super().get_index_filename()

    def generate_documents(self):
        pass

    def generate_resources(self):
        pass

    def write_resource(self, file_stream, resource):  # pragma: nocover
        pass

    def write_document(self, file_stream, document):  # pragma: nocover
        pass

    def write_index(self, file_stream):
        file_stream.write(json.dumps(self.docs_root.to_dict()))



import textwrap
from warnings import warn_explicit
from restiro.constants import (
    within_brackets_regex,
    within_parentheses_regex,
    single_word_regex,
    path_regex
)

from restiro.models import (
    FormParam, QueryParam, HeaderParam, URLParam, Resource
)

from restiro.exceptions import MissedParameter, InvalidDefinition

params_map = {
    'form': FormParam,
    'query': QueryParam,
    'head': HeaderParam,
    'url': URLParam
}


class DocstringApiResource:

    def __init__(self, docstring, filename, start_line,
                 definitions: dict = None):
        self.version = None
        self.method = None
        self.path = None
        self.group = None
        self.title = None
        self.params = []
        self.permissions = []
        self.definitions = definitions
        self.description = None
        self.filename = filename

        prepared_lines = []
        for line in docstring.split('\n'):
            # Join lines
            if line[:1] != '@':
                prepared_lines[-1] = '%s\n%s' % (prepared_lines[-1], line)
            else:
                prepared_lines.append(line)

        for index, line in enumerate(prepared_lines):
            if line.startswith('@api '):
                if self.parse_api(line, index+start_line):
                    self.title = self.title.strip()

            elif line.startswith('@apiVersion '):
                self.parse_version(line)
                self.version = self.version.strip()

            elif line.startswith('@apiGroup '):
                self.parse_group(line)
                self.group = self.group.strip()

            elif line.startswith('@apiPermission '):
                self.parse_permission(line)

            elif line.startswith('@apiDescription '):
                self.parse_description(line)
                self.description = self.description.strip()

            elif line.startswith('@apiParam '):
                self.parse_param(line, index+start_line, 'form')

            elif line.startswith('@apiQueryParam '):
                self.parse_param(line, index+start_line, 'query')

            elif line.startswith('@apiUrlParam '):
                self.parse_param(line, index+start_line, 'url')

            elif line.startswith('@apiHeadParam '):
                self.parse_param(line, index+start_line, 'head')

            elif line.startswith('@apiUse '):
                new_lines = self.parse_use_define(line, index+start_line)
                prepared_lines.extend(new_lines)

    def parse_use_define(self, line: str, index):
        name_match, name = self._get_name(line)
        name = name.strip()
        definition_lines = []
        is_found = None
        for key, define in self.definitions.items():
            if key == name:
                definition_lines = str(define.content).split('\n')
                is_found = True
        if not is_found:
            warn_explicit('There is not such apiDefine %s' % name,
                          InvalidDefinition, self.filename, index)

        return definition_lines

    @staticmethod
    def _get_type(line: str):
        try:
            type_match = next(within_brackets_regex.finditer(line))
            type_ = type_match.group()[1:-1]
        except StopIteration:
            type_match = None
            type_ = None

        return type_match, type_

    @staticmethod
    def _get_group(line: str):
        try:
            group_match = next(within_parentheses_regex.finditer(line))
            group = group_match.group()[1:-1]
        except StopIteration:
            group_match = None
            group = None

        return group_match, group

    @staticmethod
    def _get_name(line: str):
        try:
            name_match = next(single_word_regex.finditer(line))
            name = name_match.group().strip()
        except StopIteration:
            name_match = None
            name = None

        return name_match, name

    @staticmethod
    def _get_path(line: str):
        try:
            path_match = next(path_regex.finditer(line))
            path = path_match.group().strip()
        except StopIteration:
            path_match = None
            path = None

        return path_match, path

    def parse_api(self, line: str, index: int):
        type_match, type_ = self._get_type(line)
        if type_match is None:
            warn_explicit('Missed api name',
                          MissedParameter, self.filename, index)

        path_match, path = self._get_path(line)
        if path_match is None:
            warn_explicit('Missed path name',
                          MissedParameter, self.filename, index)
            return False

        type_match_span = (-1, -1) if type_match is None else type_match.span()
        path_match_span = path_match.span()

        self.method = type_
        self.path = path
        self.title = line[max(type_match_span[1], path_match_span[1]):].strip()
        return True

    def parse_permission(self, line: str):
        permissions = line.replace('@apiPermission ', '')
        permissions = permissions.split(',')
        for permission in permissions:
            self.permissions.append(permission.strip())

    def parse_group(self, line: str):
        self.group = line.replace('@apiGroup ', '')

    def parse_version(self, line: str):
        self.version = line.replace('@apiVersion ', '')

    def parse_param(self, line: str, index: int, param_type: str):
        group_match, group = self._get_group(line)
        type_match, type_ = self._get_type(line)
        name_match, name = self._get_name(line)

        if name_match is None:
            warn_explicit('Missed api parameter `name`', MissedParameter,
                          self.filename, index)
            return False

        if name.startswith('['):
            name = name[1:-1]
            optional = True
        else:
            optional = False

        group_match_span = (-1, -1) \
            if group_match is None else group_match.span()
        type_match_span = (-1, -1) if type_match is None else type_match.span()
        name_match_span = name_match.span()
        description = line[max(type_match_span[1],
                               group_match_span[1], name_match_span[1]):]
        des_lines = description.split('\n')
        des_result = ''
        for st in des_lines:
            des_result = ' '.join((des_result, st.strip())) \
                if len(des_lines) > 1 else st.strip()

        names = name.split('=')
        default = names[1] if len(names) > 1 else None

        self.params.append({
            'name': name,
            'group': group,
            'type': type_,
            'default': default,
            'description': line[max(type_match_span[1], group_match_span[1],
                                    name_match_span[1]):],
            'optional': optional,
            'param_type': param_type
        })
        return True

    def parse_description(self, line: str):
        temp_description = line.replace('@apiDescription ', '')
        # reg = re.compile('\n(?!\s*\n)([\r\t\f\v])*')
        des = temp_description.split('\n')
        t = ''
        for s in des:
            if s not in ('', ' ', '\t', '\n'):
                t = t + s.strip() + ' '
            else:
                t = t + '\n'
        self.description = t if not self.description \
            else '\n'.join((self.description, t))

    def __repr__(self):
        return '\n'.join((
            'method: %s' % self.method,
            'path: %s' % self.path,
            'title: %s' % self.title,
            'version: %s' % self.version,
            'group: %s' % self.group,
            'description: %s' % self.description,
            'permissions: %s' % self.permissions.__repr__(),
            'params: %s' % self.params.__repr__()
        ))

    def to_model(self):

        params_in_model = list(map(
            lambda param: params_map[param['param_type']](
                name=param['name'],
                description=param['description'],
                type_=param['type'],
                required=not param['optional'],
                default=param['default']
            ),
            self.params
        ))

        return Resource(
            path=self.path,
            method=self.method,
            tags=[self.group],
            display_name=self.title,
            description=self.description,
            security={'roles': self.permissions},
            params=params_in_model
        )




class ParserWarning(Warning):
    pass


class DocstringWarning(ParserWarning):
    pass


class InvalidParameter(DocstringWarning):
    pass


class InvalidDefinition(DocstringWarning):
    pass


class MissedParameter(DocstringWarning):
    pass
